from pickle import NONE
import ase.io
import os
import ccnb.bvse.Structure as Structure
import ccnb.bvse.BVAnalysis as BVAnalysis


def bv_calculation(filename,
                   moveion='Li',
                   valenceofmoveion=1,
                   resolution=0.1,
                   progress=None):
    atoms = ase.io.read(filename, store_tags=True)
    struc = Structure.Structure()
    struc.GetAseStructure(atoms)
    bvs = BVAnalysis.BVAnalysis()
    bvs.SetStructure(struc)
    bvs.SetMoveIon(moveion)
    bvs.ValenceOfMoveIon = valenceofmoveion
    bvs.SetLengthResolution(resolution)
    bvs.CaluBVSE(progress)
    bv_data = bvs.get_data()
    bvs.SaveBVSEData(os.path.splitext(filename)[0])
    bvs.SaveData(
        os.path.splitext(filename)[0] + '.bvse.grd', bv_data['BVSE'], 'BVSE')
    return bvs.Ea['BVSE']



"""


"""
import argparse
from pathlib import Path
from ccnb import cal_channel_cavd
from ccnb import get_bvse
from ccnb import get_migration_networks_voids


def get_voids(filename_CIF,
              energythreshold,
              moveion,
              mergecluster=True,
              clusterradii=0.75):
    filename_CIF = Path(filename_CIF)
    filename_CAVD = str(filename_CIF.parent.joinpath(
        filename_CIF.stem)) + ".net"
    filename_BVSE = str(filename_CIF.parent.joinpath(
        filename_CIF.stem)) + "_BVSE.npy"
    if not Path(filename_CAVD).exists():
        print('{} file no exist,\n please cavd calculate first'.format(
            filename_CAVD))
        return None
    if not Path(filename_BVSE).exists():
        print('{} file no exist,\n please bvse calculate first'.format(
            filename_BVSE))
        return None
    mn = get_migration_networks_voids(str(filename_CIF),
                                      filename_BVSE,
                                      filename_CAVD,
                                      energythreshold,
                                      moveion=moveion,
                                      mergecluster=mergecluster,
                                      clusterradii=clusterradii)
    return mn


def main():
    parser = argparse.ArgumentParser(description='ccnb 电解质材料离子输运特征计算分析程序')
    parser.add_argument('-f',
                        '--structure_file',
                        dest='struct_file',
                        type=str,
                        required=True,
                        help='结构文件名(cif格式)')
    parser.add_argument('-i',
                        '--move_ion',
                        dest='ion',
                        type=str,
                        default='Li',
                        required=True,
                        help='迁移离子类型')
    parser.add_argument('-c',
                        '--calculation_type',
                        dest='cal_type',
                        type=str,
                        default='bvse',
                        choices=['bvse', 'cavd', 'find_voids'],
                        required=True,
                        help='计算类型选择')
    bvse = parser.add_argument_group('bvse')
    bvse.add_argument('-v',
                      '--valence',
                      type=float,
                      default=1,
                      dest='valence',
                      help='迁移离子化合价')
    bvse.add_argument('-r',
                      '--resolution',
                      type=float,
                      default=0.1,
                      dest='resolution',
                      help='计算区域分辨率（单位埃）')
    cavd = parser.add_argument_group('cavd')
    cavd.add_argument('--ntol',
                      dest='ntol',
                      type=float,
                      default=0.02,
                      help='计算容限')
    cavd.add_argument('--radius_flag',
                      dest='rad_flag',
                      type=bool,
                      default=True,
                      help='几何分析时是否考虑离子半径')
    cavd.add_argument('-l',
                      '--lower',
                      dest='lower',
                      type=float,
                      default=0.5,
                      help='通道大小下限值(单位埃)')
    cavd.add_argument('-u',
                      '--upper',
                      dest='upper',
                      type=float,
                      default=1.0,
                      help='通道大小上限值(单位埃)')

    find_voids = parser.add_argument_group('find_voids')
    find_voids.add_argument('--energy',
                            dest='energy',
                            type=float,
                            default=0.5,
                            help='能量筛选上限')
    find_voids.add_argument('--cluster',
                            dest='cluster',
                            type=bool,
                            default=True,
                            help='是否进行合并团簇间隙点')
    find_voids.add_argument('-cr',
                            '--cluster_radii',
                            dest='cluster_radii',
                            type=float,
                            default=0.8,
                            help='合并团簇范围半径')
    args = parser.parse_args()
    if args.cal_type == 'cavd':
        RT = cal_channel_cavd(args.struct_file,
                              migrant=args.ion,
                              ntol=args.ntol,
                              rad_flag=args.rad_flag,
                              lower=args.lower,
                              upper=args.upper,
                              rad_dict=None)
        print(RT)
    if args.cal_type == 'bvse':
        barrier = get_bvse(args.struct_file,
                           moveion=args.ion,
                           valenceofmoveion=args.valence,
                           resolution=args.resolution)
        print(barrier)
    if args.cal_type == 'find_voids':
        mn = get_voids(filename_CIF=args.struct_file,
                       energythreshold=args.energy,
                       moveion=args.ion,
                       mergecluster=args.cluster,
                       clusterradii=args.cluster_radii)
        if mn:
            print(mn)
        else:
            print('Calculation failed!!!')


if __name__ == "__main__":
    main()



import operator
import copy
import numpy as np
import numpy.linalg as la
import networkx as nx
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from scipy.interpolate import RegularGridInterpolator
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.sites import PeriodicSite
import ccnb.mergecluster as mc
from ccnb.mergecluster import Void, Channel


class MigrationPath(object):
    """
    Calculate the  minimum energy path between endpoints, as well as the energy of each image on
    the path, the energy of bottleneck site and the migration energy barrier
    """

    def __init__(self, endpoints, phases, energy, struc, countimages=11):
        """
        :param endpoints: two-dimensional array,migration path endpoint coordinates
        :param phase: two-dimensional array.Periodic displacement vector of path
        :param energy: three-dimensional array.Bvse potential
        :param struc: pymatgen structure object
        """
        self._energy = energy
        self._struc = struc
        self._endpoints = endpoints
        self._phase = phases  # 例如[[0,0,0],[0,-1,0]]
        self._countimages = countimages
        # 路径上每点的坐标
        self._path = []
        # 路径上每点的能量
        self._path_energy = []
        # 最高点的能量
        self._max_energy = None
        self._max_position = None  # 最高点的位置
        self._min_energy = None  # 最低点的能量
        self._barrier = None  # 最高点的能量-最低点的能量
        self._energy_function = None  # 能量计算插值函数
        self.__set_energy_function()
        self.cal_path()  # 计算路径
        self.cal_energy()  # 计算路径上每点的能量

    @property
    def path(self):
        """
        Return the coordinates of each image on the MEP
        """
        return self._path

    @property
    def path_energy(self):
        """
        Return the energy of each image on the MEP. The type is a two-dimensional array.
        The first column represents the distance of each image, and the second column represents
        the energy of each image
        """
        return self._path_energy

    @property
    def energy_max(self):
        """
        :return: BVSE energy of bottleneck site
        """
        return self._max_energy

    @property
    def position_max(self):
        """
        :return: the coordinate of  bottleneck site
        """
        return self._max_position

    @property
    def energy_min(self):
        """
        :return:  the BVSE energy of the image with the lowest BVSE energy value
        """
        return self._min_energy

    @property
    def barrier(self):
        """
        :return: migration barrier
        """
        return self._barrier

    def cal_path(self):
        """
        Calculate the MEP between endpoints
        :return: Two dimensional array，every row is the coordinate of each image on the MEP
        """
        energy_shape = np.array(self._energy.shape)
        for i in range(len(self._endpoints) - 1):
            start_f = [j for j in self._endpoints[i]]
            start_f = np.mod(start_f, 1)
            end_f = [j for j in self._endpoints[i + 1]]
            end_f = np.mod(end_f, 1)
            if operator.eq(self._phase[i], [0, 0, 0]):
                start = start_f * (energy_shape - 1)
                end = end_f * (energy_shape - 1)
                self._countimages = max(
                    4,
                    int(
                        self.get_dis(self._endpoints[i],
                                     self._endpoints[i + 1]) // 0.2))
                temp_path = self.pathfind(
                    start,
                    end,
                    self._energy,
                    n_images=self._countimages,
                    dr=[
                        self._struc.lattice.a / (self._energy.shape[0] - 1),
                        self._struc.lattice.b / (self._energy.shape[1] - 1),
                        self._struc.lattice.c / (self._energy.shape[2] - 1)
                    ],
                    h=0.002,
                    k=0.17,
                    min_iter=100,
                    max_iter=10000,
                    max_tol=5e-6)
                if len(temp_path) > 3:
                    temp_path = np.round(temp_path / (energy_shape - 1), 4)
                    temp_path = np.mod(temp_path, 1.0)
                if len(self._path):
                    self._path = np.append(self._path, temp_path[1:], axis=0)
                else:
                    self._path = copy.deepcopy(temp_path)
            else:
                expan_energy2 = np.concatenate(
                    (self._energy, self._energy, self._energy), axis=0)
                expan_energy3 = np.concatenate(
                    (expan_energy2, expan_energy2, expan_energy2), axis=1)
                expan_energy = np.concatenate(
                    (expan_energy3, expan_energy3, expan_energy3), axis=2)
                start = np.array([
                    start_f[0] * (self._energy.shape[0] - 1) +
                    self._energy.shape[0], start_f[1] *
                    (self._energy.shape[1] - 1) + self._energy.shape[1],
                    start_f[2] * (self._energy.shape[2] - 1) +
                    self._energy.shape[2]
                ])
                end = end_f * (energy_shape - 1) + (
                    1 + np.array(self._phase[i])) * energy_shape
                self._countimages = max(
                    4,
                    int(
                        self.get_dis(self._endpoints[i],
                                     self._endpoints[i + 1]) // 0.2))
                temp_path = self.pathfind(
                    start,
                    end,
                    expan_energy,
                    n_images=self._countimages,
                    dr=[
                        self._struc.lattice.a / (self._energy.shape[0] - 1),
                        self._struc.lattice.b / (self._energy.shape[1] - 1),
                        self._struc.lattice.c / (self._energy.shape[2] - 1)
                    ],
                    h=0.02,
                    k=0.17,
                    min_iter=100,
                    max_iter=10000,
                    max_tol=5e-6)
                if len(temp_path) > 3:
                    temp_path = np.round(
                        (temp_path - energy_shape) / (energy_shape - 1), 4)
                    temp_path = np.mod(temp_path, 1.0)
                if len(self._path):
                    self._path = np.append(self._path, temp_path[1:], axis=0)
                else:
                    self._path = copy.deepcopy(temp_path)

    @staticmethod
    def pathfind(start,
                 end,
                 V,
                 n_images=21,
                 dr=None,
                 h=0.001,
                 k=0.17,
                 min_iter=100,
                 max_iter=10000,
                 max_tol=5e-6):
        # created by pymatgen
        if not dr:
            dr = np.array([
                1.0 / (V.shape[0] - 1), 1.0 / (V.shape[1] - 1),
                1.0 / (V.shape[2] - 1)
            ])
        else:
            dr = np.array(dr, dtype=float)
        keff = k * dr * n_images
        h0 = h
        g1 = np.linspace(0, 1, n_images)
        s0 = start
        s1 = end
        s = np.array([g * (s1 - s0) for g in g1]) + s0
        ds = s - np.roll(s, 1, axis=0)
        ds[0] = (ds[0] - ds[0])
        ls = np.cumsum(la.norm(ds, axis=1))
        ls = ls / ls[-1]
        fi = interp1d(ls, s, axis=0)
        s = fi(g1)
        # Evaluate initial distances (for elastic equilibrium)
        ds0_plus = s - np.roll(s, 1, axis=0)
        ds0_minus = s - np.roll(s, -1, axis=0)
        ds0_plus[0] = (ds0_plus[0] - ds0_plus[0])
        ds0_minus[-1] = (ds0_minus[-1] - ds0_minus[-1])
        dV = np.gradient(V)

        # Evlve string
        for step in range(0, max_iter):
            # if step > min_iter:
            # h = h0 * np.exp(-2.0 * (step - min_iter) / max_iter)  # 逐步衰减步长以减少震荡
            # else:
            h = h0
            # Calculate forces acting on string
            d = V.shape
            s0 = s
            edV = np.array([[
                dV[0][int(pt[0]) % d[0]][int(pt[1]) % d[1]][int(pt[2]) % d[2]]
                / dr[0],
                dV[1][int(pt[0]) % d[0]][int(pt[1]) % d[1]][int(pt[2]) % d[2]]
                / dr[1],
                dV[2][int(pt[0]) % d[0]][int(pt[1]) % d[1]][int(pt[2]) % d[2]]
                / dr[2]
            ] for pt in s])
            # Update according to force due to potential and string elasticity
            ds_plus = s - np.roll(s, 1, axis=0)
            ds_minus = s - np.roll(s, -1, axis=0)
            ds_plus[0] = (ds_plus[0] - ds_plus[0])
            ds_minus[-1] = (ds_minus[-1] - ds_minus[-1])
            Fpot = edV
            Fel = keff * (la.norm(ds_plus) -
                          la.norm(ds0_plus)) * (ds_plus / la.norm(ds_plus))
            Fel += keff * (la.norm(ds_minus) -
                           la.norm(ds0_minus)) * (ds_minus / la.norm(ds_minus))
            s = s - h * (Fpot + Fel)
            # Keep the coordinates of the endpoints unchanged in each iteration
            s[0] = s0[0]
            s[-1] = s0[-1]
            # Reparametrize string
            ds = s - np.roll(s, 1, axis=0)
            ds[0] = (ds[0] - ds[0])
            ls = np.cumsum(la.norm(ds, axis=1))
            ls = ls / ls[-1]
            fi = interp1d(ls, s, axis=0)
            s = fi(g1)
            tol = la.norm((s - s0) * dr) / n_images / h
            if (tol > 1e9):
                s = [[0, 0, 0]]
                break
            if (step > min_iter and tol < max_tol):
                # print("Converged at step {}".format(step))
                break
            # if (step % 100 == 0):
            # print ("Step {} - ds = {}".format(step, tol))
        return s

    def __set_energy_function(self):
        x = np.linspace(0, 1, self._energy.shape[0])
        y = np.linspace(0, 1, self._energy.shape[1])
        z = np.linspace(0, 1, self._energy.shape[2])
        self._energy_function = RegularGridInterpolator((x, y, z),
                                                        self._energy)

    def cal_point_energy(self, point):
        """
        Calculate the BVSE energy according to the discrete grid coordinates of the site point
        :param point: array，the discrete grid coordinates of the site point
        :return: float，BVSE energy
        """
        return self._energy_function(np.array(np.mod(point, 1.0)))[0]

    def get_dis(self, p1, p2):
        """
        Calculate the distance between p1 and p2 in the structure, which considers periodicity
        :param p1: One-dimensional array such as [0.11,0.22,0.33]
        :param p2:，One-dimensional
        :return: float，the distance between p1 and p2
        """
        temp_site1 = PeriodicSite('Ar', p1, self._struc.lattice)
        temp_site2 = PeriodicSite('Ar', p2, self._struc.lattice)
        dis = temp_site1.distance(temp_site2)
        return dis

    def cal_energy(self):
        """
        Calculate the BVSE energy of points on the MEP, the BVSE energy at the bottleneck and energy barriers
        """
        energy_path = []
        for point in self._path:
            point_energy = self.cal_point_energy(point)
            energy_path.append(point_energy)
        dis_path = [0.0]
        tol = 0
        for i in range(len(self._path) - 1):
            dist = self.get_dis(self._path[i], self._path[i + 1])
            tol += dist
            dis_path.append(tol)
        self._path_energy = np.zeros(shape=(len(dis_path), 2))
        for i in range(len(energy_path)):
            self._path_energy[i][1] = energy_path[i]
            self._path_energy[i][0] = dis_path[i]
        self._max_energy = max(energy_path)
        self._min_energy = min(energy_path)
        self._max_position = self._path[energy_path.index(self._max_energy)]
        self._barrier = self._max_energy - self._min_energy

    def path_to_poscar(self, filename):
        """
        Visual MEP by Poscar file
        :param filename: cif filename
        """
        struc = Structure.from_str(open(filename).read(), fmt="cif")
        struc.to(fmt='POSCAR', filename=filename.split(".")[0] + 'POSCAR')
        struc1 = Structure.from_file(filename.split(".")[0] + 'POSCAR')
        for j in range(1, len(self._path) - 1):
            struc1.insert(0, 'He', self._path[j])
        struc1.to(filename=filename.split(".")[0] + 'pathPOSCAR')


class MigrationNetwork(object):

    def __init__(self,
                 struc,
                 energy,
                 voids_dict,
                 channels_dict,
                 filename_CIF,
                 moveion='Li',
                 ismergecluster=False,
                 energythreshold=None,
                 iscalnonequalchannels=True,
                 clusterradii=0.75,
                 void_radii_threshold=[0.5, 1.0]):
        """
        Calculate the transport network,analysis transport pathways
        """
        self._moveion = moveion  # 迁移离子
        self._voids = {}  # CAVD计算的所有间隙点
        self._channels = {}  # CAVD计算的所有通道段
        self._energy = energy  # BVSE三维势场
        self._struc = struc  # Pymatgen读取的结构
        self._nonequalchannels = {}  # 以间隙为起始点的所有非等价路径
        self._mignet = None  # 间隙网络图 networkx
        self._nonequl_paths = {}  # 以晶格位为起始点的所有非等价路径
        # true、false 用于判断是否计算以间隙为起始点的所有非等价路径
        self._iscalnonequalchannels = iscalnonequalchannels
        # 用于筛选间隙和通道段的阈值
        if energythreshold:
            self._energythreshold = energythreshold + np.min(energy)
        if ismergecluster:  # 选择是否合并间隙簇
            mergevoid = mc.MergeCluster(voids_dict,
                                        channels_dict,
                                        self._struc,
                                        self._energy,
                                        filename_CIF,
                                        clusterradii=clusterradii)  # 调用合并间隙簇类
            mergevoid.to_net(filename_CIF)
            for void in mergevoid.mergedvoids:
                self._voids[void.id] = void
            for channel in mergevoid.mergedchannels:
                self._channels[(channel.start, channel.end)] = channel
        else:
            self._voids = voids_dict
            self._channels = channels_dict
        self.init_voids_channels()

    @property
    def paths_position(self):  # 返回以晶格位为起始点的所有非等价路径上每一点的坐标
        return [
            path['points_path'] for key, path in self._nonequl_paths.items()
        ]

    @property
    def paths_energy(self):  #返回以晶格位为起始点的所有非等价路径上每一点的能量
        return [
            path['energys_path'] for key, path in self._nonequl_paths.items()
        ]

    def init_voids_channels(self):
        for void_id, void in self._voids.items():
            void.energy = self.cal_point_energy(void.coord)  # 计算每一个间隙点的能量
        for channel_id, channel in self._channels.items():  # 计算每一条通道段的长度
            channel.dist = self.get_dis(self._voids[channel.start].coord,
                                        self._voids[channel.end].coord)
        if self._iscalnonequalchannels:  # 是否对每一条间隙之间的通道段计算BVSE MEP
            self.cal_nonequal_mep_between_voids()
        else:
            i = 0
            for channel_id, channel in self._channels.items(
            ):  # 计算每一条通道段的label
                if self._voids[channel.start].label < self._voids[
                        channel.end].label:
                    key = (self._voids[channel.start].label,
                           self._voids[channel.end].label,
                           round(channel.dist, 0))  # 起始点label 和长度
                else:
                    key = (self._voids[channel.end].label,
                           self._voids[channel.start].label,
                           round(channel.dist, 0))
                if key not in self._nonequalchannels.keys():
                    channel.label = i
                    self._nonequalchannels[key] = {"label": i}
                    i += 1
                else:
                    channel.label = self._nonequalchannels[key]["label"]
        if self._energythreshold:
            highenergy_voids = [
                void_id for void_id, void in self._voids.items()
                if void.energy >= self._energythreshold
            ]
            self._voids = {
                void_id: void
                for void_id, void in self._voids.items()
                if void.energy < self._energythreshold
            }
            if self._iscalnonequalchannels:
                self._channels = {
                    channel_id: channel
                    for channel_id, channel in self._channels.items()
                    if channel.start not in highenergy_voids
                    and channel.end not in highenergy_voids
                }
                self._channels = {
                    channel_id: channel
                    for channel_id, channel in self._channels.items()
                    if channel.energy < self._energythreshold
                }
            else:
                self._channels = {
                    channel_id: channel
                    for channel_id, channel in self._channels.items()
                    if channel.start not in highenergy_voids
                    and channel.end not in highenergy_voids
                }

    def cal_nonequal_mep_between_voids(self):
        # 对每一条间隙之间的通道段计算BVSE MEP
        i = 0
        for channel_id, channel in self._channels.items():
            if self._voids[channel.start].label < self._voids[
                    channel.end].label:
                key = (self._voids[channel.start].label,
                       self._voids[channel.end].label, round(channel.dist, 0))
            else:
                key = (self._voids[channel.end].label,
                       self._voids[channel.start].label,
                       round(channel.dist, 0))
            if key not in self._nonequalchannels.keys():
                mp = MigrationPath([
                    self._voids[channel.start].coord,
                    self._voids[channel.end].coord
                ], [channel.phase], self._energy, self._struc)
                channel.energy = mp.energy_max
                channel.label = i
                self._nonequalchannels[key] = {
                    "start": channel.start,
                    "end": channel.end,
                    "label": i,
                    "energy": channel.energy
                }
                i += 1
            else:
                channel.label = self._nonequalchannels[key]["label"]
                channel.energy = self._nonequalchannels[key]["energy"]
        nonequalpath = []
        for channelid, channel in self._nonequalchannels.items():
            path = {
                'start': channel["start"],
                'end': channel["end"],
                'startenergy': self._voids[channel["start"]].energy,
                'endenergy': self._voids[channel["end"]].energy,
                'bnenergy': channel['energy']
            }
            nonequalpath.append(path)
            print(path)
        return nonequalpath

    def cal_nonequl_paths(self):
        self._mignet = nx.DiGraph()
        for id_void, void in self._voids.items():
            self._mignet.add_node(void.id,
                                  label=void.label,
                                  coord=void.coord,
                                  energy=void.energy)
        for channel_id, channel in self._channels.items():
            self._mignet.add_edge(channel.start,
                                  channel.end,
                                  label=channel.label,
                                  phase=channel.phase)
        self.cal_noneqpaths_cavd()
        self.cal_noneqpaths_bvse()

    def cal_point_energy(self, point):
        # Calculate the bvse energy  according to the discrete grid coordinates of the site
        x = np.linspace(0, 1, self._energy.shape[0])
        y = np.linspace(0, 1, self._energy.shape[1])
        z = np.linspace(0, 1, self._energy.shape[2])
        interpolating_function = RegularGridInterpolator((x, y, z),
                                                         self._energy)
        point = np.mod(point, 1.0)
        return interpolating_function(np.array(point))[0]

    def get_dis(self, p1, p2):
        temp_site1 = PeriodicSite('Ar', p1, self._struc.lattice)
        temp_site2 = PeriodicSite('Ar', p2, self._struc.lattice)
        dis = temp_site1.distance(temp_site2)
        return dis

    def cal_noneqpaths_cavd(self):
        """
        calculate the label of each transport pathway
        between mobile ion lattice site
        """
        # 保存所有晶格位的类型  比如 Li1,Li2，Li3
        labels_positions = []
        # 保存所有晶格位在间隙网络中的编号，label，坐标
        positions_moveion = []
        # 字典，对所有的晶格位两两配对 key是晶格位的类型
        # 比如 （Li1,Li2） value是在间隙网络中的编号 比如（5，10）
        positions_pair = {}
        a = SpacegroupAnalyzer(self._struc, symprec=0.1)
        symm_structure = a.get_symmetrized_structure()
        # 这个循环计算positions_moveion
        # 晶格位在间隙网络中的编号，label，坐标
        for i, sites in enumerate(symm_structure.equivalent_sites):
            for site in sites:
                if site.specie.symbol == self._moveion:
                    temp_position = {}
                    temp_position['fracoord'] = site.frac_coords
                    temp_position['atom_site_label'] = site._atom_site_label
                    mini_dis = 1000
                    for void in self._voids.values():
                        dis = self.get_dis(temp_position['fracoord'],
                                           void.coord)
                        if dis < mini_dis:
                            mini_dis = dis
                            temp_position['label'] = void.label
                            temp_position['id_CAVD'] = void.id
                            if mini_dis < 0.05:
                                break
                    if temp_position['label'] not in labels_positions:
                        labels_positions.append(temp_position['label'])
                    positions_moveion.append(temp_position)
        # 这个循环计算positions_pair
        for i in range(len(positions_moveion) - 1):
            for j in range(i + 1, len(positions_moveion)):
                if positions_moveion[i]['label'] <= positions_moveion[j][
                        'label']:
                    key = (positions_moveion[i]['atom_site_label'],
                           positions_moveion[j]['atom_site_label'])
                else:
                    key = (positions_moveion[j]['atom_site_label'],
                           positions_moveion[i]['atom_site_label'])
                if positions_moveion[i]['id_CAVD'] < positions_moveion[j][
                        'id_CAVD']:
                    value = [
                        positions_moveion[i]['id_CAVD'],
                        positions_moveion[j]['id_CAVD']
                    ]
                else:
                    value = [
                        positions_moveion[j]['id_CAVD'],
                        positions_moveion[i]['id_CAVD']
                    ]
                if key not in positions_pair.keys():
                    positions_pair[key] = [value]
                else:
                    positions_pair[key].append(value)
        for pair_label, position_pairs in positions_pair.items():
            for position_pair in position_pairs:
                try:
                    voids_path = list(
                        nx.shortest_path(self._mignet,
                                         source=position_pair[0],
                                         target=position_pair[1]))
                # 计算在间隙网络中的最短路径，
                # 比如Li-Li2 结果为[1,5,7,9,2] 每一个数字表示间隙网络中的编号
                except nx.NetworkXNoPath:
                    # 如果没有路径即为空，不考虑
                    voids_path = []
                if 0 < len(voids_path) < 7:
                    temppath = {
                        'phase_path': [0, 0, 0],
                        'phases_path_segment': [],
                        'points_path': [],
                        'energys_path': [],
                        'pair_label': pair_label
                    }
                    labels_path = []  # 在该晶格位之间的每一通道段的label
                    for i in range(len(voids_path) - 1):
                        # phases_path_segment表示每一段的label，phase_path表示整条路径的label
                        labels_path.append(self._mignet[voids_path[i]][
                            voids_path[i + 1]]['label'])
                        temppath['phases_path_segment'].append(self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'])
                        temppath['phase_path'][0] += self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'][0]
                        temppath['phase_path'][1] += self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'][1]
                        temppath['phase_path'][2] += self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'][2]
                    tag0 = True  # 用于判断这条路径上每一间隙点是否为晶格位
                    voidlabels_path = [
                        self._mignet.node[voids_path[0]]["label"]
                    ]
                    for i in range(1, len(voids_path) - 1):
                        lab = self._mignet.node[voids_path[i]]["label"]
                        if lab not in labels_positions:
                            voidlabels_path.append(lab)
                        else:
                            # 如果是就不计算
                            tag0 = False
                            break
                    voidlabels_path.append(
                        self._mignet.node[voids_path[-1]]["label"])
                    # 路径上每一间隙点的label
                    temppath['voidlabels_path'] = voidlabels_path
                    # 路径上每一间隙点的编号
                    temppath['voidid_path'] = voids_path
                    # 路径上每一间隙点的坐标
                    temppath['voidcoord_path'] = [
                        self._voids[id].coord for id in voids_path
                    ]
                    temppath['barrier_path'] = None
                    if tag0 and len(labels_path) > 0:
                        # 对起始点排序，编号小的在前，
                        if labels_path[0] < labels_path[-1]:
                            key_path = tuple(labels_path)
                        else:
                            key_path = tuple(labels_path[::-1])
                        if key_path not in self._nonequl_paths.keys(
                        ):  # 判断是否在非等价路径集合里
                            self._nonequl_paths[key_path] = temppath
                        else:
                            # 和他同一label的路径
                            if operator.eq(temppath['phase_path'],
                                           [0, 0, 0]) and not (operator.eq(
                                               self._nonequl_paths[key_path]
                                               ['phase_path'], [0, 0, 0])):
                                # 如果这条路径有和他同一label的路径，但是这条和他同一label的路径是跨包的，
                                # 这条路径不跨包，就用它代替，尽可能保证选出的路径是不跨包的
                                self._nonequl_paths[tuple(
                                    labels_path)] = temppath

    def cal_noneqpaths_bvse(self):
        """
            Calculate the mep under BVSE potential
        """
        for path in self._nonequl_paths.values():
            coords_path = path['voidcoord_path']
            phases_path = path['phases_path_segment']
            mp = MigrationPath(coords_path, phases_path, self._energy,
                               self._struc)
            path['points_path'] = mp.path
            path['energys_path'] = mp.path_energy
            path['barrier_path'] = mp.barrier

    def save_data(self, filename):
        """
        Save nonequivalent transport pathways to ***_nonequalpaths.txt
        """
        with open(filename.split(".")[0] + '_nonequalpaths.txt', 'w') as f:
            i = 0
            for nonqul_id, path in self._nonequl_paths.items():
                if len(path['points_path']) > 0:
                    f.write('nonequalpath id:' + str(i) + '\n')
                    i += 1
                    f.write(path['pair_label'][0] + "  ->  " +
                            path['pair_label'][1] + "  energy barrier  " +
                            str(round(path['barrier_path'], 3)) + '\n ')
                    points = path['points_path']
                    energys = path['energys_path']
                    for j in range(len(points)):
                        f.write(
                            str(j) + "\t" + str(points[j][0]) + '  ' +
                            str(points[j][1]) + '  ' + str(points[j][2]) +
                            '\t' + str(energys[j][0]) + '  ' +
                            str(energys[j][1]) + '\n ')

    def showenergy(self, filename):
        index = 0
        for nonqul_id, path in self._nonequl_paths.items():
            if len(path['points_path']) > 0:
                xcoords = [
                    path['energys_path'][j][0]
                    for j in range(len(path['energys_path']))
                ]
                ycoords = [
                    path['energys_path'][j][1]
                    for j in range(len(path['energys_path']))
                ]
                # ycoords = [y - min(ycoords) for y in ycoords]
                poly = np.polyfit(xcoords, ycoords, deg=9)
                z = np.polyval(poly, xcoords)
                plt.figure(figsize=(8, 6))
                plt.plot(xcoords, z, linewidth=2, color="black")
                plt.xticks(fontsize=16)
                plt.yticks(fontsize=16)
                plt.scatter(xcoords, ycoords, color='k', marker='o')
                plt.xlabel("Reaction Coordinate ", fontsize=18)
                plt.ylabel("Energy", fontsize=18)
                save_path = filename.split(".")[0]+"_"+str(index) \
                            + '_' + path['pair_label'][0] \
                            + '-' + path['pair_label'][1] + '_'\
                            + '-'.join([str(i) for i in nonqul_id]) \
                            + '.svg'
                index = index + 1
                plt.savefig(save_path)
                plt.show()


def paths_to_poscar(filename_cavd,
                    filename_cif,
                    filename_bvse,
                    energythreshold=None):
    voids_list = []
    channels_list = []
    flag_p = 0
    flag_n = 0
    file = open(filename_cavd, 'r')
    for line in file.readlines():
        if 'Interstitial' in line:
            flag_p = 1
            flag_n = 0
            continue
        if 'Connection' in line:
            flag_p = 0
            flag_n = 1
            continue
        if flag_p == 1:
            line = line.split()
            if len(line) > 4:
                void = Void()
                void.id = int(line[0])
                void.label = int(line[1])
                void.coord = [
                    np.float64(line[2]),
                    np.float64(line[3]),
                    np.float64(line[4])
                ]
                void.radii = np.float64(line[5])
                voids_list.append(void)
        if flag_n == 1:
            line = line.split()
            if len(line) > 5:
                channel = Channel()
                channel.start = int(line[0])
                channel.end = int(line[1])
                channel.phase = [int(line[2]), int(line[3]), int(line[4])]
                channel.coord = [
                    np.float64(line[5]),
                    np.float64(line[6]),
                    np.float64(line[7])
                ]
                channel.radii = np.float64(line[8])
                channels_list.append(channel)
    voids = {}
    channels = {}
    for void in voids_list:
        voids[void.id] = void
    for channel in channels_list:
        if channel.start < channel.end:
            print(channel.start, channel.end)
            channels[(channel.start, channel.end)] = channel
    struc = Structure.from_str(open(filename_cif).read(), fmt="cif")
    struc.to(fmt='POSCAR', filename=filename_cif.split(".")[0] + '_POSCAR')
    struc1 = Structure.from_file(filename_cif.split(".")[0] + '_POSCAR')
    path_endpoints = []
    energy = np.load(filename_bvse)
    for channel_id, channel in channels.items():
        mp = MigrationPath(
            [voids[channel.start].coord, voids[channel.end].coord],
            [channel.phase], energy, struc)
        path = mp.path
        energy_max = mp.energy_max
        if energy_max > energythreshold:
            print(channel.start, channel.end)
        else:
            path_endpoints.append([channel.start, channel.end])
            for j in range(1, len(path) - 1):
                struc1.insert(0, 'Ne', path[j])
    for channel in path_endpoints:
        struc1.insert(0, 'He', voids[channel[0]].coord)
        struc1.insert(0, 'He', voids[channel[1]].coord)
    struc1.to(filename=filename_cif.split(".")[0] + '_POSCAR')



from pathlib import Path
from ccnb import load_struc
from ccnb import get_non_equivalent_paths_between_latticesite
from ccnb import get_migration_networks_voids
from ccnb import cal_channel_cavd
from ccnb import get_bvse
from ccnb import all_cal, configure_neb_packet

if __name__ == "__main__":
    filename_CIF = Path("xxx.cif")
    filename_CAVD = str(filename_CIF.parent.joinpath(
        filename_CIF.stem)) + ".net"
    filename_BVSE = str(filename_CIF.parent.joinpath(
        filename_CIF.stem)) + "_BVSE.npy"
    filename_CIF = str(filename_CIF)
    # Read structure from CIF file
    struc = load_struc(filename_CIF)
    print(len(struc.sites))
    # calculate interstital  network by CAVD
    RT = cal_channel_cavd(filename_CIF,
                          'Li',
                          ntol=0.02,
                          rad_flag=True,
                          lower=0.5,
                          upper=1.2,
                          rad_dict=None)
    # print(RT)
    # # calculate bond valence site example
    barrier = get_bvse(filename_CIF,
                       moveion='Li',
                       valenceofmoveion=1,
                       resolution=0.1)
    print(barrier)

    # energythreshold is used to delete path segments in the transport network that are higher than this threshold
    mep = get_migration_networks_voids(filename_CIF,
                                       filename_BVSE,
                                       filename_CAVD,
                                       energythreshold=0.4,
                                       moveion='Li',
                                       mergecluster=False)
    mep = get_non_equivalent_paths_between_latticesite(filename_CIF,
                                                       filename_BVSE,
                                                       filename_CAVD,
                                                       energythreshold=-1.3,
                                                       moveion='Li')
    print(mep)
    configure_neb_packet(filename_CIF, mep.paths_position)
    all_cal(filename_CIF,
            moveion='Li',
            valenceofmoveion=1,
            energythreshold=1.4)



from pymatgen.io.vasp.inputs import Incar, Kpoints, Potcar
from pymatgen.core import Structure
from functools import reduce
import sys, os
from pymatgen.core.sites import Site

from ccnb.vasp_parse import parse_potcar
from ccnb.path_poscar import path_poscar


class vasp_inputs(object):

    def __init__(self, struc):
        self.incar_params = {
            'pp': 'pbe',
            'istart': 0,
            'icharg': 2,
            'lwave': False,
            'lcharg': False,
            'lreal': 'auto',
            'prec': 'accurate',
            'gga': 'pe',
            'encut': 400,  #!!!!
            'ediff': 1e-4,  #!!!!
            'ediffg': -0.05,  #!!!!
            'nelm': 60,  #!!!!
            'nelmin': 2,  #!!!!
            'nsw': 1000,  #!!!!
            'isif': 2,  #!!!!
            'sigma': 0.2,  #!!!!
            'npar': 2,
            'ibrion': 2,
            'ismear': -5
        }
        self.neb_incar = {
            'spring': -5,
            'lclimb': True,
            'iopt': 1,  #!!!!
            'potim': 0,
            'ichain': 0,
            'ibrion': 3,  #!!!!
            'ismear': 0,
            'algo': 'fast'
        }
        self.vasp_path = 'vasp'
        self.struc = struc
        self.kpts = ()
        self.symbols = []
        self.set_kpts()
        self.set_symbols()
        self.dir = ''

    def set_incar(self, dict):
        for k in dict:
            self.incar_params[k] = dict[k]

    def set_lsf(self, vasp_path):
        self.vasp_path = vasp_path

    def set_kpts(self):
        self.kpts = ((int(round(30 / self.struc.lattice.a)),
                      int(round(30 / self.struc.lattice.b)),
                      int(round(30 / self.struc.lattice.c))), )

    def set_symbols(self):
        func = lambda x, y: x if y in x else x + [y]
        species = reduce(func, [
            [],
        ] + self.struc.species)
        for i in range(len(species)):
            self.symbols.append(str(species[i]))

    def potcar(self):
        try:
            potc = Potcar(symbols=self.symbols)  #functional=u'PBE' default
            potc.write_file(str(self.dir) + '/POTCAR')
        except OSError as errorinfo:
            print(str(errorinfo))
            return 0

    def incar(self, neb):
        file = open(str(self.dir) + '/POTCAR')
        p = parse_potcar(file.read())
        self.set_incar({'encut': 1.5 * p.get_max_enmax()})
        file.close()

        if neb == True:
            self.incar_params.update(self.neb_incar)
        incar = Incar(self.incar_params)
        incar.write_file(str(self.dir) + '/INCAR')

    def kpoints(self):
        kp = Kpoints(comment=u'Automatic mesh',
                     kpts=self.kpts,
                     style=Kpoints.supported_modes.Monkhorst)
        kp.write_file(str(self.dir) + '/KPOINTS')

    def lsf(self, neb):
        n = 16
        if neb:
            n = 16 * self.incar_params['images']
        filename = os.path.split(self.dir)[1]
        file = open(str(self.dir) + '/vasp.lsf', 'wb')
        file.write('#!/bin/sh\n'.encode('ascii'))
        file.write('#BSUB -cwd.\n'.encode('ascii'))
        file.write('#BSUB -e %J.err\n'.encode('ascii'))
        file.write('#BSUB -o %J.out\n'.encode('ascii'))
        file.write('#BSUB -q priority\n'.encode('ascii'))
        file.write(('#BSUB -n ' + str(n) + '\n').encode('ascii'))
        file.write(('#BSUB -J ' + str(filename) + '\n').encode('ascii'))
        file.write('#BSUB -x\n'.encode('ascii'))
        file.write(
            'ncpus=`cat $LSB_DJOB_HOSTFILE | wc -l ` \n'.encode('ascii'))
        file.write(('mpirun  -hostfile $LSB_DJOB_HOSTFILE  -np  ${ncpus} ' +
                    self.vasp_path + '\n').encode('ascii'))
        file.close()

    def inputs(self, dir, neb=False):
        self.dir = dir
        self.potcar()
        self.incar(neb)
        self.kpoints()
        self.lsf(neb)
        path_poscar(dir + '/POSCAR1', dir + '/POSCAR2', dir + '/path.cif')


#path includes the start and end sites
def interpolate(dir, struc1, struc2, path):
    nimages = len(path) - 1
    images = struc1.interpolate(struc2, nimages, True)
    dir = os.dirname()

    i = 0
    for struc in images:
        struc.translate_sites(0,
                              path[i] - struc.sites[0].frac_coords,
                              frac_coords=True,
                              to_unit_cell=True)
        num = ('%02d' % i)
        if not os.path.exists(dir + '/' + num):
            os.mkdir(dir + '/' + num)
        struc.to(filename=dir + '/' + num + '/POSCAR')
        i = i + 1


# struc = Structure.from_file('test/POSCAR')
# vi = vasp_inputs('Li',struc)
# vi.set_incar({'nelect':128})
# vi.inputs('test',False)
# vi.path_poscar([1,1,1])



import zipfile, os


def dfs_get_zip_file(input_path, result):
    files = os.listdir(input_path)
    for file in files:
        if os.path.isdir(input_path + '/' + file):
            dfs_get_zip_file(input_path + '/' + file, result)
        else:
            result.append(input_path + '/' + file)


def zip_path(input_path, output_name):
    f = zipfile.ZipFile(output_name, 'w', zipfile.ZIP_DEFLATED)
    filelists = []
    dfs_get_zip_file(input_path, filelists)
    for file in filelists:
        f.write(file)
    f.close()
    return output_name



#from pymatgen.io.vasp.inputs import PotcarSingle
import pymatgen.io.vasp.inputs as vin
import re


class parse_potcar(object):
    parse_functions = {
        "LULTRA": vin._parse_bool,
        "LUNSCR": vin._parse_bool,
        "LCOR": vin._parse_bool,
        "LPAW": vin._parse_bool,
        "EATOM": vin._parse_float,
        "RPACOR": vin._parse_float,
        "POMASS": vin._parse_float,
        "ZVAL": vin._parse_float,
        "RCORE": vin._parse_float,
        "RWIGS": vin._parse_float,
        "ENMAX": vin._parse_float,
        "ENMIN": vin._parse_float,
        "EMMIN": vin._parse_float,
        "EAUG": vin._parse_float,
        "DEXC": vin._parse_float,
        "RMAX": vin._parse_float,
        "RAUG": vin._parse_float,
        "RDEP": vin._parse_float,
        "RDEPT": vin._parse_float,
        "QCUT": vin._parse_float,
        "QGAM": vin._parse_float,
        "RCLOC": vin._parse_float,
        "IUNSCR": vin._parse_int,
        "ICORE": vin._parse_int,
        "NDATA": vin._parse_int,
        "VRHFIN": vin._parse_string,
        "LEXCH": vin._parse_string,
        "TITEL": vin._parse_string,
        "STEP": vin._parse_list,
        "RRKJ": vin._parse_list,
        "GGA": vin._parse_list
    }

    def __init__(self, data):
        search_lines = re.findall(r"(?s)(parameters from PSCTR are:"
                                  r".*?END of PSCTR-controll parameters)",
                                  data)  #.group(2)

        self.all_keywords = []
        for i in range(len(search_lines)):
            keywords = {}
            for key, val in re.findall(r"(\S+)\s*=\s*(.*?)(?=;|$)",
                                       search_lines[i],
                                       flags=re.MULTILINE):
                try:
                    keywords[key] = self.parse_functions[key](val)
                except KeyError:
                    warnings.warn("Ignoring unknown variable type %s" % key)
            self.all_keywords.append(keywords)

    def get_max_enmax(self):
        max_enmax = 0
        for i in range(len(self.all_keywords)):
            enmax = self.all_keywords[i]['ENMAX']
            if max_enmax < enmax:
                max_enmax = enmax
        return max_enmax


# p=PotcarSingle(file.read())
# print(p.enmax)



#author:cst
#
#input files:
#1.POSCAR(parent relaxed structure)
#2.POSCAR1 POSCAR2 (relaxed start and end structure)
#
#parameter:
#1.nelect for neb calculation
#2.migration_ion, default value 'Li'
#
#out file:
#neb packages for vasp:include POSCAR1 POSCAR2 (start and end structure nedd to relax)

import numpy as np
import sys, os
from pymatgen.core import Structure

from ccnb.zip_paths import zip_path
from ccnb.vasp_inputs import vasp_inputs


class neb_packages(object):
    #dir: the poscar dir
    #filename: structure name,such as icsd_5
    def init(self, filename, ion='Li'):
        self.nelect = 128  ######notice this
        self.migration_ion = ion
        #self.paths=[] #all paths from start to end
        self.filename = filename
        self.dir = os.path.dirname(filename)
        struc = Structure.from_file(self.dir + '/POSCAR')
        self.vi = vasp_inputs(struc)
        self.vi.set_incar({'nelect': self.nelect})

    def deal_path_periodicity(self, path):
        new_path = []
        num = 0
        #print('path:',path)
        for coord in path:
            coord = np.mod(coord, 1)
            #if(num%2==0):
            new_path.append(coord)
            #num=num+1
        #print('new_path:',new_path)
        return new_path

    def from_file(self, path_file):
        self.paths = np.load(path_file)

    def from_list(self, paths):
        self.paths = paths

    def data_parse(self):
        if not os.path.exists(self.dir + '/paths'):
            os.mkdir(self.dir + '/paths')
        num = 0
        for path in self.paths:
            # 处理坐标周期性问题
            p = self.deal_path_periodicity(path)
            dir = self.dir + '/paths/path_' + str(num)
            # dir=self.paths_dir+'/path_'+str(num)
            if self.judge_site(p[0], p[-1], dir):
                continue
            struc = Structure.from_file(self.dir + '/POSCAR')
            struc.to(filename=dir + '/POSCAR')
            for i in range(1, len(p) - 1):
                struc.insert(0, 'He', p[i])
            struc.to(fmt='cif', filename=dir + '/path.cif')
            struc1 = Structure.from_file(dir + '/POSCAR_base')
            struc1.insert(
                0, self.migration_ion,
                path[0])  # at the first site insert the interval point
            struc1.to(filename=dir + "/POSCAR1")
            struc2 = Structure.from_file(dir + '/POSCAR_base')
            struc2.insert(
                0, self.migration_ion,
                path[-1])  # at the first site insert the interval point
            struc2.to(filename=dir + "/POSCAR2")
            self.vi.set_incar({'images': len(p) - 2})

            self.vi.inputs(dir, True)
            num = num + 1
        zip_path(self.dir + '/paths', self.filename + '_neb_paths.zip')

    # product the base poscar for insert site
    def judge_site(self, p1, p2, dir):
        start = None
        end = None
        # print('start:%s', p1)
        # print('end:%s', p2)
        struc = Structure.from_file(self.dir + '/POSCAR')
        struc.insert(0, self.migration_ion, p2)
        struc.insert(0, self.migration_ion, p1)
        dis = struc.get_distance(0, 1)
        # print(dir, 'dis:', dis)
        # if (dis > 6): #if distance greater than 5, then can not participate in neb calculation
        # print('dis too long!!!!!!!!')
        # return 1
        for i in range(2, len(struc.sites)):
            # print('%s:%s', i,struc.sites[i].frac_coords)
            dis1 = struc.get_distance(i, 0)
            dis2 = struc.get_distance(i, 1)
            # print('dis:',dis1,dis2)
            if dis1 < 0.5:
                start = i
                # print('dis1:',dis1)
            if dis2 < 0.5:
                end = i
                # print('dis2:',dis2)
        # print('start:%s', start)
        # print('end:%s',end)
        if start:
            struc.sites.pop(start)
            if end:
                if end > start:
                    struc.sites.pop(end - 1)
                elif end < start:
                    struc.sites.pop(end)
        elif end:
            struc.sites.pop(end)
        else:
            print('no site')
            return 1
        struc.sites.pop(0)
        struc.sites.pop(0)
        if not os.path.exists(dir):
            os.mkdir(dir)
        struc.to(filename=dir + '/POSCAR_base')



import operator
import math
import numpy as np
import networkx as nx
from scipy.interpolate import RegularGridInterpolator
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.sites import PeriodicSite
import ccnb.mergecluster as mc
from ccnb.bvse_cavd import MigrationPath


class MigrationNetworkLatticeSites(object):

    def __init__(self,
                 struc,
                 energy,
                 voids_dict,
                 channels_dict,
                 filename_CIF,
                 moveion='Li',
                 ismergecluster=False):
        self._moveion = moveion
        self._voids = {}
        self._channels = {}
        self._energy = energy
        self._struc = struc
        self._mignet = None
        self._sites = {}
        self._nonequalchannels = {}
        self._nonequl_paths = {}
        self.all_paths = {}
        self._id_barrier_phase = {}
        if ismergecluster:
            mergevoid = mc.MergeCluster(voids_dict,
                                        channels_dict,
                                        self._struc,
                                        self._energy,
                                        filename_CIF,
                                        clusterradii=0.5)
            mergevoid.to_net(filename_CIF)
            for void in mergevoid.mergedvoids:
                self._voids[void.id] = void
            for channel in mergevoid.mergedchannels:
                self._channels[(channel.start, channel.end)] = channel
        else:
            self._voids = voids_dict
            self._channels = channels_dict
        self.init_voids_channels()

    def init_voids_channels(self):
        for void_id, void in self._voids.items():
            void.energy = self.cal_point_energy(void.coord)
        for channel_id, channel in self._channels.items():
            channel.dist = self.get_dis(self._voids[channel.start].coord,
                                        self._voids[channel.end].coord)
        i = 0
        for channel_id, channel in self._channels.items():
            if self._voids[channel.start].label < self._voids[
                    channel.end].label:
                key = (self._voids[channel.start].label,
                       self._voids[channel.end].label, round(channel.dist, 1))
            else:
                key = (self._voids[channel.end].label,
                       self._voids[channel.start].label,
                       round(channel.dist, 1))
            if key not in self._nonequalchannels.keys():
                channel.label = i
                self._nonequalchannels[key] = {"label": i}
                i += 1
            else:
                channel.label = self._nonequalchannels[key]["label"]

    def cal_paths(self):
        self._mignet = nx.DiGraph()
        for id_void, void in self._voids.items():
            self._mignet.add_node(void.id,
                                  label=void.label,
                                  coord=void.coord,
                                  energy=void.energy)
        for channel_id, channel in self._channels.items():
            self._mignet.add_edge(channel.start,
                                  channel.end,
                                  label=channel.label,
                                  phase=channel.phase)
        self.cal_paths_cavd()
        self.cal_noneqpaths_bvse()
        for key_path, path in self.all_paths.items():
            phase2 = [0, 0, 0]
            for i in range(3):
                if path['phase_path'][i] != 0:
                    phase2[i] = -1 * path['phase_path'][i]
            if path['voidlabels_path'][0] < path['voidlabels_path'][-1]:
                self._id_barrier_phase[(path['voidid_path'][0], path['voidid_path'][-1])] = \
                    [self._nonequl_paths[path['key_path']]['barrier_begin_to_end'], path['phase_path']]
                self._id_barrier_phase[(path['voidid_path'][-1], path['voidid_path'][0])] = \
                    [self._nonequl_paths[path['key_path']]['barrier_end_to_begin'], phase2]
            else:
                self._id_barrier_phase[(path['voidid_path'][0], path['voidid_path'][-1])] = \
                    [self._nonequl_paths[path['key_path']]['barrier_end_to_begin'], path['phase_path']]
                self._id_barrier_phase[(path['voidid_path'][-1], path['voidid_path'][0])] = \
                    [self._nonequl_paths[path['key_path']]['barrier_begin_to_end'], phase2]
        for site in self._sites.values():
            site["neighbor"] = []
            site['barriers'] = []
            site['phases'] = []
            for endpoints, barrier_phase in self._id_barrier_phase.items():
                if endpoints[0] == site['id_CAVD']:
                    site["neighbor"].append(endpoints[1])
                    site['barriers'].append(barrier_phase[0])
                    site['phases'].append(barrier_phase[1])

    def fac2cart(self, coord):
        """
        分数坐标转换成笛卡尔坐标
        """
        return np.dot(coord, self._struc.lattice.matrix)

    def cart2fac(self, coord):
        """
        笛卡尔坐标转换成分数坐标
        """
        return np.dot(coord, np.linalg.inv(self._struc.lattice.matrix))

    def cal_point_energy(self, point):
        # Calculate the bvse energy  according to the discrete grid coordinates of the site
        x = np.linspace(0, 1, self._energy.shape[0])
        y = np.linspace(0, 1, self._energy.shape[1])
        z = np.linspace(0, 1, self._energy.shape[2])
        interpolating_function = RegularGridInterpolator((x, y, z),
                                                         self._energy)
        point = [math.modf(i)[0] for i in point]
        for i in range(len(point)):
            if point[i] < 0.0:
                point[i] += 1
        return interpolating_function(np.array(point))[0]

    def get_dis(self, p1, p2):
        temp_site1 = PeriodicSite('Ar', p1, self._struc.lattice)
        temp_site2 = PeriodicSite('Ar', p2, self._struc.lattice)
        dis = temp_site1.distance(temp_site2)
        return dis

    def cal_paths_cavd(self):
        """
        calculate the label of each transport pathway between mobile ion lattice site
        """
        labels_positions = []
        positions_moveion = []
        positions_pair = {}
        a = SpacegroupAnalyzer(self._struc, symprec=0.1)
        symm_structure = a.get_symmetrized_structure()
        for i, sites in enumerate(symm_structure.equivalent_sites):
            for site in sites:
                if site.specie.symbol == self._moveion:
                    temp_position = {}
                    temp_position['atom_site_label'] = site._atom_site_label
                    mini_dis = 1000
                    for void in self._voids.values():
                        dis = self.get_dis(site.frac_coords, void.coord)
                        if dis < mini_dis:
                            mini_dis = dis
                            temp_position['label'] = void.label
                            temp_position['id_CAVD'] = void.id
                            if mini_dis < 0.01:
                                break
                    temp_position['fracoord'] = self._voids[
                        temp_position['id_CAVD']].coord
                    temp_position['cartcoord'] = self.fac2cart(
                        temp_position['fracoord'])
                    if temp_position['label'] not in labels_positions:
                        labels_positions.append(temp_position['label'])
                    positions_moveion.append(temp_position)
                    print(temp_position)
                    self._sites[temp_position['id_CAVD']] = temp_position
        for i in range(len(positions_moveion) - 1):
            for j in range(i + 1, len(positions_moveion)):
                key = (positions_moveion[j]['atom_site_label'],
                       positions_moveion[i]['atom_site_label'])
                if positions_moveion[i]['label'] < positions_moveion[j][
                        'label']:
                    value = [
                        positions_moveion[i]['id_CAVD'],
                        positions_moveion[j]['id_CAVD']
                    ]
                else:
                    value = [
                        positions_moveion[j]['id_CAVD'],
                        positions_moveion[i]['id_CAVD']
                    ]
                if key not in positions_pair.keys():
                    positions_pair[key] = [value]
                else:
                    positions_pair[key].append(value)
        for pair_label, position_pairs in positions_pair.items():
            for position_pair in position_pairs:
                try:
                    voids_path = list(
                        nx.shortest_path(self._mignet,
                                         source=position_pair[0],
                                         target=position_pair[1]))
                except nx.NetworkXNoPath:
                    voids_path = []
                if 0 < len(voids_path) < 7:
                    temppath = {
                        'phase_path': [0, 0, 0],
                        'phases_path_segment': [],
                        'points_path': [],
                        'energys_path': [],
                        'pair_label': pair_label
                    }
                    labels_path = []
                    for i in range(len(voids_path) - 1):
                        labels_path.append(self._mignet[voids_path[i]][
                            voids_path[i + 1]]['label'])
                        temppath['phases_path_segment'].append(self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'])
                        temppath['phase_path'][0] += self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'][0]
                        temppath['phase_path'][1] += self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'][1]
                        temppath['phase_path'][2] += self._mignet[
                            voids_path[i]][voids_path[i + 1]]['phase'][2]
                    tag0 = True
                    voidlabels_path = [
                        self._mignet.node[voids_path[0]]["label"]
                    ]
                    for i in range(1, len(voids_path) - 1):
                        lab = self._mignet.node[voids_path[i]]["label"]
                        if lab not in labels_positions:
                            voidlabels_path.append(lab)
                        else:
                            tag0 = False
                            break
                    voidlabels_path.append(
                        self._mignet.node[voids_path[-1]]["label"])
                    temppath['voidlabels_path'] = voidlabels_path
                    temppath['voidid_path'] = voids_path
                    temppath['voidcoord_path'] = [
                        self._voids[id].coord for id in voids_path
                    ]
                    temppath['barrier_path'] = None
                    if tag0:
                        if labels_path[0] < labels_path[-1]:
                            key_path = tuple(labels_path)
                        else:
                            key_path = tuple(labels_path[::-1])
                        temppath['key_path'] = key_path
                        self.all_paths[(voids_path[0],
                                        voids_path[-1])] = temppath
                        if key_path not in self._nonequl_paths.keys():
                            self._nonequl_paths[key_path] = temppath
                        else:
                            if operator.eq(temppath['phase_path'], [0, 0, 0]) and \
                               not (operator.eq(self._nonequl_paths[key_path]['phase_path'], [0, 0, 0])):
                                self._nonequl_paths[tuple(
                                    labels_path)] = temppath

    def cal_noneqpaths_bvse(self):
        for path in self._nonequl_paths.values():
            coords_path = path['voidcoord_path']
            phases_path = path['phases_path_segment']
            mp = MigrationPath(coords_path, phases_path, self._energy,
                               self._struc)
            path['points_path'] = mp.path
            path['energys_path'] = mp.path_energy
            energys = []
            for i in range(len(path['energys_path'])):
                energys.append(path['energys_path'][i][1])
            i_max = 0
            for i in range(len(energys)):
                if energys[i] == mp.energy_max:
                    i_max = i
                    break
            if path['voidlabels_path'][0] < path['voidlabels_path'][0]:
                path['barrier_begin_to_end'] = mp.energy_max - min(
                    energys[:i_max])
                path['barrier_end_to_begin'] = mp.energy_max - min(
                    energys[i_max:])
            else:
                path['barrier_begin_to_end'] = mp.energy_max - min(
                    energys[i_max:])
                path['barrier_end_to_begin'] = mp.energy_max - min(
                    energys[:i_max])

    def save_data(self, filename):
        """
        Save nonequivalent transport pathways to ***_nonequalpaths.txt
        """
        with open(filename.split(".")[0] + '_mc.txt', 'w') as f:
            f.write('Lattice_basis_vector ')
            f.write('\n')
            f.write(str(self._struc.lattice))
            f.write('\n\n')
            for site in self._sites.values():
                print(site)
                f.write('Site_number ' + str(site['id_CAVD']) + '\n')
                f.write('Cartesian_coordinates ' + str(site['cartcoord'][0]) +
                        ' ' + str(site['cartcoord'][1]) + ' ' +
                        str(site['cartcoord'][2]) + '\n')
                f.write('Normalized_coordinates ' + str(site['fracoord'][0]) +
                        ' ' + str(site['fracoord'][1]) + ' ' +
                        str(site['fracoord'][2]) + '\n')
                f.write('Neighbor_sites')
                for nb in site['neighbor']:
                    f.write(' ' + str(nb))
                f.write('\n')
                f.write('Neighbor_sites_phase')
                for ph in site['phases']:
                    f.write(' ' + str(ph[0]) + ' ' + str(ph[1]) + ' ' +
                            str(ph[2]) + ',')
                f.write('\n')
                f.write('Barrier')
                for ba in site['barriers']:
                    f.write(' ' + str(ba))
                f.write('\n')
                f.write('Site_types ' + str(site['atom_site_label']) + '\n')
                f.write('\n')



from pymatgen.core.structure import Structure
from pymatgen.core.sites import PeriodicSite
from scipy.interpolate import RegularGridInterpolator
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
import numpy as np
import math
import copy
import numpy.linalg as la
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from scipy.interpolate import make_interp_spline as spline


def pathfind(start, end, V, n_images=21, dr=None, h=0.1, k=0.17, min_iter=100, max_iter=10000, max_tol=5e-6):
    if not dr:
        dr = np.array([1.0 / (V.shape[0] - 1), 1.0 / (V.shape[1] - 1), 1.0 / (V.shape[2] - 1)])
    else:
        dr = np.array(dr, dtype=float)
    keff = k * dr * n_images
    h0 = h
    g1 = np.linspace(0, 1, n_images)
    s0 = start
    s1 = end
    s = np.array([g * (s1 - s0) for g in g1]) + s0
    ds = s - np.roll(s, 1, axis=0)
    ds[0] = (ds[0] - ds[0])
    ls = np.cumsum(la.norm(ds, axis=1))
    ls = ls / ls[-1]
    fi = interp1d(ls, s, axis=0)
    s = fi(g1)
    # Evaluate initial distances (for elastic equilibrium)
    ds0_plus = s - np.roll(s, 1, axis=0)
    ds0_minus = s - np.roll(s, -1, axis=0)
    ds0_plus[0] = (ds0_plus[0] - ds0_plus[0])
    ds0_minus[-1] = (ds0_minus[-1] - ds0_minus[-1])
    dV = np.gradient(V)
    for step in range(0, max_iter):
        if step > min_iter:
            h = h0 * np.exp(-2.0 * (step - min_iter) / max_iter)
        else:
            h = h0
        # Calculate forces acting on string
        d = V.shape
        s0 = s
        edV = np.array([[dV[0][int(pt[0]) % d[0]][int(pt[1]) % d[1]][int(pt[2]) % d[2]] / dr[0],

                         dV[1][int(pt[0]) % d[0]][int(pt[1]) % d[1]][int(pt[2]) % d[2]] / dr[1],

                         dV[2][int(pt[0]) % d[0]][int(pt[1]) % d[1]][int(pt[2]) % d[2]] / dr[2]] for pt in s])

        # Update according to force due to potential and string elasticity
        ds_plus = s - np.roll(s, 1, axis=0)
        ds_minus = s - np.roll(s, -1, axis=0)
        ds_plus[0] = (ds_plus[0] - ds_plus[0])
        ds_minus[-1] = (ds_minus[-1] - ds_minus[-1])
        Fpot = edV
        Fel = keff * (la.norm(ds_plus) - la.norm(ds0_plus)) * (ds_plus / la.norm(ds_plus))
        Fel += keff * (la.norm(ds_minus) - la.norm(ds0_minus)) * (ds_minus / la.norm(ds_minus))
        s = s - h * (Fpot + Fel)
        s[0] = s0[0]
        s[-1] = s0[-1]
        ds = s - np.roll(s, 1, axis=0)
        ds[0] = (ds[0] - ds[0])
        ls = np.cumsum(la.norm(ds, axis=1))
        ls = ls / ls[-1]
        fi = interp1d(ls, s, axis=0)
        s = fi(g1)
        tol = la.norm((s - s0) * dr) / n_images / h
        if (tol > 1e9):
            s = [[0, 0, 0]]
            break
        if (step > min_iter and tol < max_tol):
            break
    return s


def get_dis(struc,p1, p2):
    temp_site1 = PeriodicSite('Ar', p1, struc.lattice)
    temp_site2 = PeriodicSite('Ar', p2, struc.lattice)
    dis = temp_site1.distance(temp_site2)
    return dis


def calpointenergy(energy, point):
    x = np.linspace(0, energy.shape[0] - 1, energy.shape[0])
    y = np.linspace(0, energy.shape[1] - 1, energy.shape[1])
    z = np.linspace(0, energy.shape[2] - 1, energy.shape[2])
    interpolating_function = RegularGridInterpolator((x, y, z), energy)
    return interpolating_function(np.array(point))


def pathenergy(energy,struc, p):
    energy_path = []
    for point_of_p in p:
        point_temp = [0.0,0.0,0.0]
        for i in range(len(point_temp)):
            if point_of_p[i] < 0:
                 point_temp[i] = (point_of_p[i]+1)*(energy.shape[i]-1)
            elif point_of_p[i] <= 1:
                point_temp[i] = point_of_p[i] * (energy.shape[i]-1)
            else:
                point_temp[i] = (point_of_p[i]-1) * (energy.shape[i] - 1)
        energy_point_temp = calpointenergy(energy, point_temp)
        energy_path.append(energy_point_temp)
    dis_path = [0.0]
    tol = 0
    for i in range(len(p)-1):
        dist = get_dis(struc,p[i], p[i+1])
        tol += dist
        dis_path.append(tol)
    for i in range(len(dis_path)):
        dis_path[i] = (dis_path[i]/dis_path[-1]) *100
    migs = np.zeros(shape=(len(dis_path), 2))
    for i in range(len(energy_path)):
        migs[i][1] = energy_path[i]
        migs[i][0] = dis_path[i]
    return migs


def calpath(endpoints,struc, energy,count_images = 11):
    paths = []
    for i in range(len(endpoints)-1):
        start_f = endpoints[i]
        end_f = endpoints[i+1]
        start = np.array([start_f[0] * (energy.shape[0] - 1), start_f[1] * (energy.shape[1] - 1),
                          start_f[2] * (energy.shape[2] - 1)])
        end = np.array(
            [end_f[0] * (energy.shape[0] - 1), end_f[1] * (energy.shape[1] - 1), end_f[2] * (energy.shape[2] - 1)])
        #count_images = max(5, int(get_dis(struc,start_f, end_f) // 0.2))
        p= pathfind(start, end, energy, n_images=count_images,
                     dr=[struc.lattice.a / (energy.shape[0] ),
                         struc.lattice.b / (energy.shape[1] ),
                         struc.lattice.c / (energy.shape[2] )], h=0.01, k=0.17, min_iter=100,
                     max_iter=50000, max_tol=5e-6)
        print(p)
        for p1 in p:
            p1[0] = round(p1[0] / (energy.shape[0] - 1), 5)
            p1[1] = round(p1[1] / (energy.shape[1] - 1), 5)
            p1[2] = round(p1[2] / (energy.shape[2] - 1), 5)
        if i == 0:
            paths = copy.deepcopy(p)
        else:
            paths = np.append(paths, p[1:], axis=0)
    return paths


def paths_2_cif(paths, p_e, filename):
    with open(filename.split(".")[0] + 'path_energy_message.txt', 'w') as f:
        for i in range(len(paths)):
            f.write(str(p_e[i][0]) + '  ' + str(p_e[i][1]) + '\n ')
        for i in range(len(paths)):
            f.write(str(paths[i][0]) + '  ' + str(paths[i][1]) + ' ' + str(paths[i][1]) + '\n ')
    struc1 = Structure.from_file(filename)
    #struc.to(fmt='POSCAR', filename=filename.split(".")[0] + '_POSCAR')
    #struc1 = Structure.from_file(filename.split(".")[0] + '_POSCAR')
    for j in range(1, len(paths) - 1):
        struc1.insert(0, 'He', paths[j])
    struc1.to(fmt="CIF", filename=filename.split(".")[0] + 'path')


def showpathsenergy(p_e):
    xcoords = []
    ycoords = []
    for j in range(len(p_e)):
        xcoords.append(p_e[j][0])
        ycoords.append(p_e[j][1])

    x_new = np.linspace(min(xcoords), max(xcoords), 100)
    y_smooth = spline(xcoords, ycoords)(x_new)
    plt.figure(figsize=(6, 4))
    plt.scatter(xcoords, ycoords, color='k', marker='o')
    plt.plot(x_new, y_smooth,linewidth=2, color ='k')
    plt.xticks(fontsize=15)
    plt.yticks(fontsize=15)
    plt.xlabel("Migration path(%) ", fontsize=20)
    plt.ylabel("Energy (eV)", fontsize=20)
    plt.show()


if __name__ == "__main__":
    filename_BVSE = 'D:\\temp\\zhouzy\\JF1-new-2_BVSE.npy'
    filename_CIF = 'D:\\temp\\zhouzy\\JF1-new-2-hb.cif'
    struc = Structure.from_file(filename_CIF)
    energy = np.load(filename_BVSE)
    energy = energy - np.amin(energy)
    #Li5->Li4->Li1->L2->Li3
    endpoints = [[0.754 , 0.353,  0.361],
                 [0.746 , 0.147 , 0.639]]
    p =calpath(endpoints, struc,energy,count_images=10)
    #count_images 是每段路径插入点的个数
    print(p)  #输出坐标
    p_e = pathenergy(energy, struc, p)
    print(p_e)  #输出能量
    showpathsenergy(p_e)





from monty.io import zopen
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.spatial import cKDTree
import networkx as nx
import ase.spacegroup as spg
from pymatgen.core.sites import PeriodicSite
from cavd.local_environment import CifParser_new


class Void(object):

    def __init__(self):
        self.id = None
        self.label = None
        self.coord = None
        self.radii = None
        self.energy = None


class Channel(object):

    def __init__(self):
        self.start = None
        self.end = None
        self.phase = None
        self.coord = None
        self.radii = None
        self.dist = None
        self.label = None
        self.energy = None


class MergeCluster(object):

    def __init__(self,
                 voids_dict,
                 channels_dict,
                 structure,
                 bvse,
                 filename_cif,
                 clusterradii=0.5):
        self._struc = structure
        self._energy = bvse
        self._voids = voids_dict
        self._channels = channels_dict
        self._mergedvoids = []
        self._mergedchannels = []
        self._clusterradii = clusterradii
        self._clusters = []
        self.init_voids_channels()
        self.find_clusters()
        self.handle_voidsandchannels()
        self.cal_void_label(filename_cif)

    @property
    def mergedvoids(self):
        """
        Return to all merged voids
        """
        return self._mergedvoids

    @property
    def mergedchannels(self):
        """
        Return to all merged channel segments
        """
        return self._mergedchannels

    def get_absolute_dis(self, p1, p2):
        """
        Calculate the distance between two sites without considering periodicity
        :param p1: fractional coordinates of site p1，such as[0.5，0.5，0.5]
        :param p2: fractional coordinates of site p2
        :return: float, the distance between two sites
        """
        coord1 = np.array(self.fac2cart(p1))
        coord2 = np.array(self.fac2cart(p2))
        diff = np.subtract(coord1, coord2)
        return np.linalg.norm(diff)

    def get_period_dis(self, p1, p2):
        """
        Calculate the distance between two sites with considering periodicity
        :param p1: fractional coordinates of site p1，such as[0.5，0.5，0.5]
        :param p2: fractional coordinates of site p2
        :return:  the distance between two sites
        """
        temp_site1 = PeriodicSite('Ar', p1, self._struc.lattice)
        temp_site2 = PeriodicSite('Ar', p2, self._struc.lattice)
        dis = temp_site1.distance(temp_site2)
        return dis

    def fac2cart(self, coord):
        """
        Converting fractional coordinates to Cartesian coordinates
        """
        return np.dot(coord, self._struc.lattice.matrix)

    def cart2fac(self, coord):
        """
        Converting Cartesian coordinates to fractional coordinates
        """
        return np.dot(coord, np.linalg.inv(self._struc.lattice.matrix))

    def cal_point_energy(self, point):
        """
        Calculate the bond valence site energy of a site in crystal structure
        :param point: fractional coordinates of a site
        :return: float,the bond valence site energy of a site
        """
        x = np.linspace(0, 1, self._energy.shape[0])
        y = np.linspace(0, 1, self._energy.shape[1])
        z = np.linspace(0, 1, self._energy.shape[2])
        interpolating_function = RegularGridInterpolator((x, y, z),
                                                         self._energy)
        for i in range(len(point)):
            if point[i] < 0.0:
                point[i] += 1
            if point[i] > 1.0:
                point[i] -= 1
        return interpolating_function(np.array(point))[0]

    def init_voids_channels(self, radii_threadhold=0.5):
        """
        Initialize interstices and channel segments,delete too small interstices and channel segments
        :param radii_threadhold: float, filter threshold
        """
        for void_id, void in self._voids.items():
            void.energy = self.cal_point_energy(void.coord)
        small_voids = [
            void_id for void_id, void in self._voids.items()
            if void.radii < radii_threadhold
        ]
        self._voids = {
            void_id: void
            for void_id, void in self._voids.items()
            if void.radii >= radii_threadhold
        }
        self._channels = {
            channel_id: channel
            for channel_id, channel in self._channels.items() if
            channel.start not in small_voids and channel.end not in small_voids
        }
        self._channels = {
            channel_id: channel
            for channel_id, channel in self._channels.items()
            if channel.radii >= radii_threadhold
        }

    def find_clusters(self):
        """
        find each clusters in interstital network
        """
        coords = []
        pair_voids = []
        voids_list = [void for void_id, void in self._voids.items()]
        for void in voids_list:
            coords.append(self.fac2cart(void.coord))
        coord_tree = cKDTree(coords)
        all_pair_voids = [
            i for i in coord_tree.query_pairs(r=self._clusterradii)
        ]
        for i in all_pair_voids:
            pair_voids.append([voids_list[i[0]].id, voids_list[i[1]].id])
        if len(pair_voids) > 0:
            graph_clusters = nx.Graph()
            for e in pair_voids:
                graph_clusters.add_edge(e[0], e[1])
            queue_clusters = []
            for sc in nx.connected_component_subgraphs(graph_clusters):
                queue_clusters.append(list(sc.nodes))
            while queue_clusters:
                subv_in = queue_clusters.pop(0)
                subv_out = []
                temp_coord = [0.0, 0.0, 0.0]
                for subv in subv_in:
                    temp_coord[0] += self.fac2cart(self._voids[subv].coord)[0]
                    temp_coord[1] += self.fac2cart(self._voids[subv].coord)[1]
                    temp_coord[2] += self.fac2cart(self._voids[subv].coord)[2]
                centre_coord = self.cart2fac([
                    temp_coord[0] / len(subv_in), temp_coord[1] / len(subv_in),
                    temp_coord[2] / len(subv_in)
                ])
                for i in range(len(subv_in)):
                    if self.get_period_dis(
                            self._voids[subv_in[i]].coord,
                            centre_coord) > self._clusterradii + 0.2:
                        subv_out.append(subv_in[i])
                for subv in subv_out:
                    subv_in.remove(subv)
                self._clusters.append(subv_in)
                if len(subv_out) > 1:
                    pair_subvout = []
                    for i in range(len(subv_out)):
                        for j in range(i + 1, len(subv_out)):
                            if self.get_period_dis(
                                    self._voids[subv_out[i]].coord,
                                    self._voids[subv_out[j]].coord
                            ) < self._clusterradii:
                                pair_subvout.append([subv_out[i], subv_out[j]])
                    if len(pair_subvout) > 0:
                        graph_subvout = nx.Graph()
                        for e in pair_subvout:
                            graph_subvout.add_edge(e[0], e[1])
                        for sc in nx.connected_component_subgraphs(
                                graph_subvout):
                            queue_clusters.append(list(sc.nodes))
        print("Number of clusters", len(self._clusters))
        print(self._clusters)

    def handle_voidsandchannels(self):
        """
        handle clusters
        """
        mignet = nx.Graph()
        for void_id, void in self._voids.items():
            mignet.add_node(void.id,
                            label=void.label,
                            coord=void.coord,
                            radii=void.radii)
        for e_id, e in self._channels.items():
            if e.start < e.end:
                phase1 = e.phase
                phase2 = [0, 0, 0]
                for i in range(3):
                    if phase1[i] != 0:
                        phase2[i] = -1 * phase1[i]
                mignet.add_edge(e.start,
                                e.end,
                                phase1=phase1,
                                phase2=phase2,
                                coord1=e.coord,
                                radii1=e.radii,
                                dist1=e.dist)
        if len(self._clusters) > 0:
            for i in range(len(self._clusters)):
                minenergy = 20000
                centervoid_id = self._clusters[i][0]
                for void_id in self._clusters[i]:
                    if minenergy > self._voids[void_id].energy:
                        centervoid_id = void_id
                        minenergy = self._voids[void_id].energy
                center_void = Void()
                center_void.id = centervoid_id
                center_void.label = self._voids[centervoid_id].label
                center_void.coord = self._voids[centervoid_id].coord
                center_void.radii = self._voids[centervoid_id].radii
                center_void.energy = self._voids[centervoid_id].energy
                tempedges = []
                nearvoids = []
                for nearvoid in list(mignet.adj[centervoid_id].keys()):
                    if nearvoid not in self._clusters[i]:
                        nearvoids.append(nearvoid)
                        if centervoid_id < nearvoid:
                            start = centervoid_id
                            end = nearvoid
                        else:
                            end = centervoid_id
                            start = nearvoid
                        tempedges.append({
                            "from":
                            start,
                            "to":
                            end,
                            "phase1":
                            mignet[centervoid_id][nearvoid]['phase1'],
                            "phase2":
                            mignet[centervoid_id][nearvoid]['phase2'],
                            "coord1":
                            mignet[centervoid_id][nearvoid]['coord1'],
                            "radii1":
                            mignet[centervoid_id][nearvoid]['radii1'],
                            "dist1":
                            mignet[centervoid_id][nearvoid]['dist1']
                        })
                for id in self._clusters[i]:
                    if id != center_void.id:
                        for nearvoid in list(mignet.adj[id].keys()):
                            if nearvoid not in self._clusters[
                                    i] and nearvoid not in nearvoids:
                                if centervoid_id < nearvoid:
                                    start = centervoid_id
                                    end = nearvoid
                                else:
                                    end = centervoid_id
                                    start = nearvoid
                                if id < nearvoid:
                                    ph_cen_nearvoid = mignet[id][nearvoid][
                                        'phase1']
                                    ph_nearvoid_cen = mignet[id][nearvoid][
                                        'phase2']
                                else:
                                    ph_cen_nearvoid = mignet[id][nearvoid][
                                        'phase2']
                                    ph_nearvoid_cen = mignet[id][nearvoid][
                                        'phase1']
                                if centervoid_id < nearvoid:
                                    ph1 = ph_cen_nearvoid
                                    ph2 = ph_nearvoid_cen
                                else:
                                    ph2 = ph_cen_nearvoid
                                    ph1 = ph_nearvoid_cen
                                tempedges.append({
                                    "from":
                                    start,
                                    "to":
                                    end,
                                    "phase1":
                                    ph1,
                                    "phase2":
                                    ph2,
                                    "coord1":
                                    mignet[id][nearvoid]['coord1'],
                                    "radii1":
                                    mignet[id][nearvoid]['radii1'],
                                    "dist1":
                                    mignet[id][nearvoid]['dist1']
                                })
                for void in self._clusters[i]:
                    mignet.remove_node(void)
                mignet.add_node(center_void.id,
                                label=center_void.label,
                                coord=center_void.coord,
                                radii=center_void.radii)
                for e in tempedges:
                    mignet.add_edge(e['from'],
                                    e['to'],
                                    phase1=e['phase1'],
                                    phase2=e['phase2'],
                                    coord1=e['coord1'],
                                    radii1=e['radii1'],
                                    dist1=e['dist1'])
        for nd in mignet.nodes():
            tempvoid = Void()
            tempvoid.id = nd
            tempvoid.label = mignet.nodes[nd]['label']
            tempvoid.coord = mignet.nodes[nd]['coord']
            tempvoid.radii = mignet.nodes[nd]['radii']
            self._mergedvoids.append(tempvoid)
        for edge in mignet.edges():
            tempchannel1 = Channel()
            tempchannel2 = Channel()
            tempchannel1.start = edge[0]
            tempchannel1.end = edge[1]
            tempchannel2.end = edge[0]
            tempchannel2.start = edge[1]
            if edge[0] < edge[1]:
                tempchannel1.phase = mignet[edge[0]][edge[1]]["phase1"]
                tempchannel2.phase = mignet[edge[0]][edge[1]]["phase2"]
            else:
                tempchannel1.phase = mignet[edge[0]][edge[1]]["phase2"]
                tempchannel2.phase = mignet[edge[0]][edge[1]]["phase1"]
            tempchannel1.coord = mignet[edge[0]][edge[1]]["coord1"]
            tempchannel2.coord = mignet[edge[0]][edge[1]]["coord1"]
            tempchannel1.radii = mignet[edge[0]][edge[1]]["radii1"]
            tempchannel2.radii = mignet[edge[0]][edge[1]]["radii1"]
            tempchannel1.dist = mignet[edge[0]][edge[1]]["dist1"]
            tempchannel2.dist = mignet[edge[0]][edge[1]]["dist1"]
            self._mergedchannels.append(tempchannel1)
            self._mergedchannels.append(tempchannel2)

    @staticmethod
    def tag_sites(sitesym, scaled_positions, symprec=1e-5):
        scaled = np.around(np.array(scaled_positions, ndmin=2), 8)
        scaled %= 1.0
        scaled %= 1.0
        np.set_printoptions(suppress=True)
        tags = -np.ones((len(scaled), ), dtype=int)
        tagdis = 100 * np.ones((len(scaled), ), dtype=float)
        rot, trans = spg.spacegroup.parse_sitesym(sitesym)
        siteskdTree = cKDTree(scaled)
        for i in range(len(scaled)):
            if tags[i] == -1:
                curpos = scaled[i]
                sympos = np.dot(rot, curpos) + trans
                sympos %= 1.0
                sympos %= 1.0
                sympos = np.unique(np.around(sympos, 8), axis=0)
                min_dis, min_ids = siteskdTree.query(sympos, k=1)
                select = min_dis < symprec
                select_ids = min_ids[select]
                tags[select_ids] = i
                tagdis[select_ids] = min_dis[select]
        return tags, tagdis

    def cal_void_label(self, filename_cif, symprec=0.1):
        """
        Classification of interstices based on symmetry
        """
        with zopen(filename_cif, "rt") as f:
            input_string = f.read()
        parser = CifParser_new.from_string(input_string)
        sitesym = parser.get_sym_opt()
        voids_positions = []
        for void in self.mergedvoids:
            voids_positions.append(void.coord)
        tags, tagdis = self.tag_sites(sitesym, voids_positions, symprec)
        for i in range(len(tags)):
            self.mergedvoids[i].label = tags[i]

    def to_net(self, filename):
        """
        Save the merged interstices and channel fragments to a net file
        :param filename: output will be written to a file
        """
        with open(filename.split(".")[0] + '_mergecluster.net', 'w') as f:
            f.write('Interstitial table:\n')
            for void in self.mergedvoids:
                f.write(
                    str(void.id) + "\t" + str(void.label) + "\t " +
                    str(void.coord[0]) + " " + str(void.coord[1]) + " " +
                    str(void.coord[2]) + "\t " + str(void.radii) + "\n")
            f.write('Connection table:\n')
            for channel in self.mergedchannels:
                f.write(
                    str(channel.start) + "\t " + str(channel.end) + "\t " +
                    str(int(channel.phase[0])) + " " +
                    str(int(channel.phase[1])) + " " +
                    str(int(channel.phase[2])) + "\t " +
                    str(channel.coord[0]) + " " + str(channel.coord[1]) + " " +
                    str(channel.coord[2]) + "\t " + str(channel.radii) + "\n")

    def get_clusters(self):
        """
        Find all interstice clusters in a periodic unit cell
        :return: list, all interstice clusters
        """
        clusters = []
        for cluster in self._clusters:
            cluster_temp = []
            for void_id in cluster:
                cluster_temp.append({
                    "void_id": void_id,
                    "void_coord": self._voids[void_id].coord
                })
            clusters.append(cluster_temp)
        return clusters



import os
from pathlib import Path
from struct import unpack
from monty.io import zopen
import numpy as np
import pymatgen
from pymatgen.core import Structure
from pymatgen.core import SETTINGS
from cavd.local_environment import CifParser_new
from ccnb.bvse_cavd import MigrationNetwork
from ccnb.mergecluster import Void, Channel
from ccnb.neb_packages import neb_packages
from ccnb.bvse import bv_calculation
from ccnb.cavd_channel import cal_channel_cavd
from rich.progress import Progress


class Progressbar:

    def __init__(self):
        self.progress = Progress()
        self.task = None

    def set_task(self, workstr: str, total):
        self.task = self.progress.add_task(workstr, total)

    def updatetask(self, i: int):
        self.progress.update(self.task, advance=0.9)
        print('bvse compute have finished {0:.2f}%'.format(i))


def get_channel_cavd(filename,
                     migrant,
                     ntol=0.02,
                     rad_flag=True,
                     lower=0,
                     upper=10.0,
                     rad_dict=None):
    """
    calculate interstitial network and Voronoi network by CAVD
    :param filename: CIF filename
    :param migrant: mobile ion
    :param lower: the lower threshold  of ion migration
                  to select the suitable interstitial network
    :param upper: the upper threshold  of ion migration
                  to select the suitable interstitial network
    :return: RT value, NET file for saving
             interstitial network and Voronoi network
    """
    conn_val = cal_channel_cavd(filename,
                                migrant,
                                ntol=ntol,
                                rad_flag=rad_flag,
                                lower=lower,
                                upper=upper,
                                rad_dict=rad_dict)
    return conn_val


def get_bvse(filename_cif,
             moveion='Li',
             valenceofmoveion=1,
             resolution=0.1,
             progress=None):
    """
    calculate BVSE landscape
    :param filename_cif: cif filename
    :param moveion:  move ion
    :param valenceofmoveion: valence of move ion
    :param resolution: resolution for calculating BVSE landsacpe
    :return: migration energy barrier, numpy Binary.
             Pgrid file for saving and visualization BVSE landscape
    """
    prgbar = Progressbar()
    prgbar.set_task('bvse computing progress...', total=100)
    barrier = bv_calculation(filename_cif,
                             moveion=moveion,
                             valenceofmoveion=valenceofmoveion,
                             resolution=resolution,
                             progress=prgbar.updatetask)
    return barrier


def load_struc(filename_cif):
    """
    read structure from CIF file
    :param filename_cif: cif filename
    :return: Structure object defined by pymatgen
    """
    with zopen(filename_cif, "rt") as f:
        input_string = f.read()
    parser = CifParser_new.from_string(input_string)
    structure = parser.get_structures(primitive=False)[0]
    return structure


def load_bvse_from_npy(filename_bvse):
    """
    Read BVSE potential from numpy binary file
    :param filename_bvse: numpy binary filename
    :return: Three dimensional array
    """
    energy_landscape = np.load(filename_bvse)
    return energy_landscape


def load_bvse_from_pgrid(filename_bvse):
    '''
     Read BVSE potential from pgrid file
    :param filename_bvse: potential filename
    :return: three dimensional array
    '''
    binfile = open(filename_bvse, 'rb')
    data = binfile.read(112)
    data1 = binfile.read(12)
    num = unpack('3i', data1)
    total_num = str(num[0] * num[1] * num[2]) + 'f'
    data2 = binfile.read(28)
    data_num = binfile.read()
    num_energy = unpack(total_num, data_num)
    energy_landscape = np.array(num_energy).reshape((num[0], num[1], num[2]))
    return energy_landscape


def load_voids_channels_from_file(filename_cavd):
    """
    Read interstices and channel segments from net file calculated by cavd
    :param filename_cavd: NET filename
    :return: dict, interstices and channel segments
    """
    voids_dict = {}
    channels_dict = {}
    flag_p = 0
    flag_n = 0
    file = open(filename_cavd, 'r')
    for line in file.readlines():
        if 'Interstitial' in line:
            flag_p = 1
            flag_n = 0
            continue
        if 'Connection' in line:
            flag_p = 0
            flag_n = 1
            continue
        if flag_p == 1:
            line = line.split()
            if len(line) > 3:
                void = Void()
                void.id = int(line[0])
                void.label = int(line[1])
                void.coord = [
                    np.float64(line[2]),
                    np.float64(line[3]),
                    np.float64(line[4])
                ]
                void.radii = np.float64(line[5])
                voids_dict[void.id] = void
        if flag_n == 1:
            line = line.split()
            if len(line) > 4:
                channel = Channel()
                channel.start = int(line[0])
                channel.end = int(line[1])
                channel.phase = [int(line[2]), int(line[3]), int(line[4])]
                channel.coord = [
                    np.float64(line[5]),
                    np.float64(line[6]),
                    np.float64(line[7])
                ]
                channel.radii = np.float64(line[8])
                channels_dict[(channel.start, channel.end)] = channel
    return voids_dict, channels_dict


def get_non_equivalent_paths_between_latticesite(filename_CIF,
                                                 filename_BVSE,
                                                 filename_CAVD,
                                                 energythreshold,
                                                 moveion='Li'):
    """
    get nonequivalent transport pathways between lattice sites
    :param filename_CIF:  CIF filename
    :param filename_BVSE: BVSE landscape numpy binary
    :param filename_CAVD: net filename
    :param energythreshold: energythreshold is used to delete path segments 
           in the transport network
    that are higher than this threshold
    :param moveion: move ion
    :return: the fractional coordinates of each image on the MEP
    """
    voids, channels = load_voids_channels_from_file(filename_CAVD)
    struc = load_struc(filename_CIF)
    energy = load_bvse_from_npy(filename_BVSE)
    mn = MigrationNetwork(struc,
                          energy,
                          voids,
                          channels,
                          filename_CIF,
                          moveion=moveion,
                          ismergecluster=True,
                          energythreshold=energythreshold,
                          iscalnonequalchannels=False)
    mn.cal_nonequl_paths()
    mn.save_data(filename_CIF)
    mn.showenergy(filename_CIF)
    return mn
    # return mn.paths_position，return mn.paths_position


def get_migration_networks_voids(filename_CIF,
                                 filename_BVSE,
                                 filename_CAVD,
                                 energythreshold,
                                 moveion='Li',
                                 mergecluster=True,
                                 clusterradii=0.75):
    """
    获得迁移网络中的间隙点
    """
    voids, channels = load_voids_channels_from_file(filename_CAVD)
    struc = Structure.from_file(filename_CIF)
    energy = load_bvse_from_npy(filename_BVSE)
    mn = MigrationNetwork(struc,
                          energy,
                          voids,
                          channels,
                          filename_CIF,
                          moveion=moveion,
                          ismergecluster=mergecluster,
                          energythreshold=energythreshold,
                          iscalnonequalchannels=False,
                          clusterradii=clusterradii)
    for void in mn._voids.values():
        struc.insert(0, 'He', void.coord)
    struc.to(fmt='cif', filename=filename_CIF + '_voids.cif')
    return mn


def get_non_equivalent_paths_between_voids(filename_CIF,
                                           filename_BVSE,
                                           filename_CAVD,
                                           energythreshold=None,
                                           moveion='Li'):
    voids, channels = load_voids_channels_from_file(filename_CAVD)
    struc = load_struc(filename_CIF)
    energy = load_bvse_from_npy(filename_BVSE)
    mn = MigrationNetwork(struc,
                          energy,
                          voids,
                          channels,
                          filename_CIF,
                          moveion=moveion,
                          ismergecluster=True,
                          energythreshold=energythreshold,
                          iscalnonequalchannels=True)
    return mn.cal_nonequal_mep_between_voids()


def configure_neb_packet(filename_CIF, mep, moveion="Li"):
    struc1 = load_struc(filename_CIF)
    path_cif_file = Path(filename_CIF).parent.joinpath('POSCAR')

    struc1.to(fmt='POSCAR', filename=str(path_cif_file))
    file_prefix = filename_CIF.split('.')[0]
    #SETTINGS['PMG_VASP_PSP_DIR'] = os.path.abspath("..") + '\ccnb\psp_resources'
    n = neb_packages()
    n.init(file_prefix, moveion)
    n.from_list(mep)
    n.data_parse()


def all_cal(filename_cif,
            moveion='Li',
            valenceofmoveion=1,
            energythreshold=None):
    bv_calculation(filename_cif,
                   moveion=moveion,
                   valenceofmoveion=valenceofmoveion,
                   resolution=0.1)
    cal_channel_cavd(filename_cif,
                     moveion,
                     ntol=0.02,
                     rad_flag=True,
                     lower=0,
                     upper=10.0,
                     rad_dict=None)
    filename_BVSE = filename_cif.split('.')[0] + '.npy'
    filename_CAVD = filename_cif.split('.')[0] + '_origin.net'
    meps = get_non_equivalent_paths_between_latticesite(filename_cif,
                                                        filename_BVSE,
                                                        filename_CAVD,
                                                        energythreshold,
                                                        moveion=moveion)
    configure_neb_packet(filename_cif, meps, moveion=moveion)



import os, re
from monty.io import zopen
import numpy as np
from cavd.netio import *
from cavd.channel import Channel
from cavd.netstorage import AtomNetwork, connection_values_list
from cavd.local_environment import CifParser_new, LocalEnvirCom
from cavd.get_Symmetry import get_labeled_vornet
from pymatgen.core. structure import Structure

def cal_channel_cavd(filename,
                     migrant,
                     ntol=0.02,
                     rad_flag=True,
                     lower=0,
                     upper=10.0,
                     rad_dict=None):
    with zopen(filename, "rt") as f:
        input_string = f.read()
    parser = CifParser_new.from_string(input_string)
    stru = parser.get_structures(primitive=False)[0]
    #stru = Structure.from_file(filename)
    #species = [str(sp).replace("Specie ", "") for sp in stru.species]
    #elements = [re.sub('[^a-zA-Z]', '', sp) for sp in species]
    elements = stru.symbol_set
    #symm_number, symm_sybol = parser.get_symme()
    sitesym = parser.get_sym_opt()
    if migrant not in elements:
        raise ValueError(
            "The input migrant ion not in the input structure! Please check it."
        )
    effec_radii, migrant_radius, migrant_alpha, nei_dises, coordination_list = LocalEnvirCom(
        stru, migrant)
    radii = {}
    if rad_flag:
        if rad_dict:
            radii = rad_dict
        else:
            radii = effec_radii
    print(radii)
    atmnet = AtomNetwork.read_from_RemoveMigrantCif(filename, migrant, radii,
                                                    rad_flag)
    vornet, edge_centers, fcs, faces = atmnet.perform_voronoi_decomposition(
        True, ntol)
    # add_fcs_vornet = vornet.add_facecenters(faces)
    prefixname = filename.replace(".cif", "")
    symprec = 0.01
    sym_vornet, voids = get_labeled_vornet(vornet, sitesym, symprec)
    writeNETFile(prefixname + "_origin.net", atmnet, sym_vornet)
    channels = Channel.findChannels2(vornet, atmnet, lower, upper,
                                     prefixname + ".net")
    Channel.writeToVESTA(channels, atmnet, prefixname)
    conn_val = connection_values_list(prefixname + ".resex", sym_vornet)
    return conn_val




from cavd.local_environment import CifParser_new
from ccnb.bvse_cavd import MigrationNetwork
from ccnb.mergecluster import Void, Channel
from ccnb.neb_packages import neb_packages
from ccnb.bvse import bv_calculation
from ccnb import (load_struc,load_bvse_from_npy,
          load_voids_channels_from_file,get_bvse)
from ccnb.cavd_channel import cal_channel_cavd
from ccnb.MigrationNetworkLatticeSites import MigrationNetworkLatticeSites


if __name__ == "__main__":
    filename_CIF = 'tzw_2021032612573832.cif'
    # calculate interstital  network by CAVD
    #RT =cal_channel_cavd(filename_CIF, 'Li', ntol=0.02, rad_flag=True, lower=0.4, upper=10.0, rad_dict=None)
    #print(RT)
    # calculate bond valence site example
    barrier = get_bvse(filename_CIF, moveion='Li', valenceofmoveion=1, resolution=0.1)
    print(barrier)
    #filename_CAVD = "icsd_221\\icsd_221.net"
    #filename_BVSE = "icsd_221\\icsd_221_BVSE.npy"
    #voids, channels = load_voids_channels_from_file(filename_CAVD)
    #struc = load_struc(filename_CIF)
    #energy = load_bvse_from_npy(filename_BVSE)
    #print(voids)
    #mn = MigrationNetworkLatticeSites(struc, energy, voids, channels, filename_CIF, moveion="Li")
    #mn.cal_paths()
    #mn.save_data(filename_CIF)






from pymatgen.core import Structure
import os, sys

# from vasp_inputs import interpolate


#product the neb packages for one path
#posc1:file of POSCAR1POSCAR1 posc2:file of POSCAR3 posc_path:file of POSCAR_path
def path_poscar(posc1, posc2, posc_path):
    struc1 = Structure.from_file(posc1)
    struc2 = Structure.from_file(posc2)
    s = Structure.from_file(posc_path)

    path = []
    path.append(struc1.sites[0].frac_coords)
    for site in s.sites:
        if site.specie.symbol == 'He':
            path.append(site.frac_coords)
    path.append(struc2.sites[0].frac_coords)

    nimages = len(path) - 1
    images = struc1.interpolate(struc2, nimages, True)
    dir = os.path.dirname(posc_path)

    i = 0
    for struc in images:
        struc.translate_sites(0,
                              path[i] - struc.sites[0].frac_coords,
                              frac_coords=True,
                              to_unit_cell=True)
        num = ('%02d' % i)
        if not os.path.exists(dir + '/' + num):
            os.mkdir(dir + '/' + num)
        struc.to(filename=dir + '/' + num + '/POSCAR')
        i = i + 1

    # interpolate(dir, struc1, struc2, path)


#path_poscar(sys.argv[0], sys.argv[1], sys.argv[2])



from setuptools import setup, find_packages


def readme():
    with open('README.en.md', encoding='utf-8') as f:
        content = f.read()
    return content


install_requires = [
    "ase==3.15.0",
    "pymatgen==2022.0.10",
    "cavd==0.5.4",
]

setup(name="ccnb",
      version="0.2.2",
      author='He Bing',
      author_email='bhe@t.shu.edu.cn',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://gitee.com/shuhebing/CCNB',
      packages=find_packages(),
      include_package_data=True,
      install_requires=install_requires,
      python_requires='>=3.7,<3.10',
      entry_points={
          'console_scripts': [
              'ccnb = ccnb:main.main',
          ],
      })
